/********************************************************************/

#include "Gen.h"

/********************************************************************/

// The global version string as set by the build system.
static const CHAR *projectVersionString = VERSION;

// The path to the binary.
static char projectBinPath[PATH_MAX]="/call/genInit/in/main";

/********************************************************************/

// Return 1 if the string in trail trails at the end
// of the string (is followed by a null) or 0 otherwise.
static int stringTrails (char *first , char *trail)
{
    // See if trail occurs in the string
    char *temp = strstr (first , trail);

    // If it does, and the character right after it is a null byte, then it trails.
    if (temp != NULL && first[(temp - first) + strlen (trail)] == '\0')
        return 1;

    // No trail!
    return 0;
}

/********************************************************************/

// This function takes a string that represents the argv[0] that was passed
// to the main of the running program, and a buffer (of size PATH_MAX), and
// attempts to determine the location of the binary.
//
// To do that it, it checks for path information in argv[0]. If it finds some, it
// can use realpath to resolve that to a full directory path, based on the current
// directory. If not, we search the path to find a binary.
//
// The runpath argument (which should be a text buffer of size PATH_MAX) will be
// filled with the directory of the binary on return.
static void getBinaryPath (char *argv0, char *runPath)
{
    char tempPath[PATH_MAX];
    struct stat sb;

    // First, make a copy of the argv0 we were given
    strcpy (tempPath, argv0);

    // If there is some path information in our temporay path, then we can use
    // realpath to resolve it to an absolute pathname. Otherwise, we have do to
    // more work.
    if (strrchr (tempPath, '/') == NULL)
    {
        // Pull the PATH and HOME out of the environment. The path thing is a little annoying,
        // because we want the easy to a temporary pointer into the path, with the safeness of
        // a copy of it. So we make a copy, then alias a pointer to it.
        char copyOfPath[8192];
        char *path;
        char *home = getenv ("HOME");
        strncpy (copyOfPath , getenv ("PATH") , sizeof (copyOfPath) - 1);
        copyOfPath[sizeof (copyOfPath) - 1] = '\0';
        path = copyOfPath;

        // If there is some path information, then we can go to town with it
        while (path)
        {
            // Get the location of the first path in the path list, by finding the
            // first path seperation character. If we find one, we will clobber it over,
            // to terminate the path there.
            char *PATHspot = strchr (path, ':');
            if (PATHspot)
                *PATHspot = '\0';

            // If We found a path location, and there's still path left, then set up a
            // temporary filename using that directory, to see if the executable is
            // in there. If none of these criteria is met, then search the current
            // directory instead.
            if ( (!PATHspot && *path) || (PATHspot > (path+1)) )
            {
                // If the path segment starts with a tilde, then we want to expand that.
                // Otherwise, just use the path as is
                if ((*path == '~') && home != NULL )
                    sprintf (tempPath, "%s/%s/%s", home , path + 1 , argv0);
                else
                    sprintf (tempPath , "%s/%s" , path , argv0);
            } else {
                sprintf (tempPath , "./%s" , argv0);
            }

            // Advance the path now, to be one character past where we clobered, so it is 
            // at the start of the next path segment (or the end of the string).
            path = PATHspot + 1;

            // Stat the file, to see if we can access it. This makes sure there's not a directory
            // or other special file with the same name as our binary on the path
            if ((access (tempPath, X_OK) == 0) && (stat (tempPath, &sb) == 0) && S_ISREG (sb.st_mode))
                break;

            // If we're at the end of the path and we haven't found the program, then we have no choice
            // but to just drop through.
            if (PATHspot == NULL)
            {
                strcpy (tempPath, argv0);
                break;
            }
        }
    }

    // At this point, tempPath has as much path information as we have managed to get. It could be
    // a relative path, if the user invoked the program like that (i.e. ../../Bin/QuoteServer, for
    // example). Or, it could be a directory that was found in the path. Calling realpath will make
    // the path absolute, cleaning it up along the way.
    if (realpath (tempPath, runPath) == NULL)
    {
        perror ("realpath failed, defaulting DataMillRoot to current directory");
        strcpy (runPath , "./");
        return;
    }

    // Now runPath is the complete path, with no relative bits. Copy it back to tempPath, so we can
    // run realpath on it again.
    strcpy (tempPath , runPath);

    // Clobber out the stuff after the last / in the string, so we just have some path information.
    if (strrchr (tempPath , '/'))
        *(strrchr (tempPath , '/') + 1) = '\0';

    // Check to see if the bin directory is pased onto the binary, and if so, undo it.
    if (stringTrails (tempPath , "/bin/"))
        strcat (tempPath , "../");

    // Call realpath one more time, to clean those up.
    if (realpath (tempPath, runPath) == NULL)
    {
        perror ("realpath failed, defaulting DataMillRoot to current directory");
        strcpy (runPath , "./");
        return;
    }
}

/********************************************************************/

// Initialize the Gen library. This sets up various variables for us
// so that other calls can be made to get their values later. If we
// are supporting ObjC code, this also creates a default NSAutorelease
void genInit (int argc , char **argv)
{
    // Initialize the project binary path.
    getBinaryPath (argv[0] , projectBinPath);

    // Now that we're done, call genInitObjC to setup ObjC specific
    // things. 
    genInitObjC (argc , argv);
}

/********************************************************************/

// This reverses everything that genInit does
void genShutdown (void)
{
    // To start with, shutdown any ObjC specific things now, before we
    // do anything else, in case it depends on something we're about to
    // clean up.
    genShutdownObjC ();

    // Now finish the rest of the shutdown.
}

/********************************************************************/

// Determine the project root (directory under the directory that the current
// binary is being run in) and return it as a pointer to a static buffer.
const CHAR *getProjectRootCSTRING (void)
{
    return projectBinPath;
}

/********************************************************************/

// Determine the project version (specified in Makefiles/ReleaseNumber) and
// return it as a pointer to a static buffer.
const CHAR *getProjectVersionCSTRING (void)
{
    // Return the string that was set by the build system when we were compiled.
    return projectVersionString;
}

/********************************************************************/


