/********************************************************************/

#import "Gen.h"

/********************************************************************/

// This function gets invoked by gdb when you use the "po" or "print-object" command
// on an Objective-C object. It's supposed to return the string to print. However, we
// don't want to pollute the heap with temporary objects created by some object's
// description method, in case we're debugging something memory related. So we'll
// print the thing ourselves, and then return back an empty string, so that we can
// clean up all objects that were created by the call.
//
// Note that this may require a hacked up version of gdb, since most versions
// (at least prior to 5.3) don't include support for debugging ObjC.
// 
// This isn't included in a header file anywhere, since we just need the symbol
// to exist in the executable for gdb to find it.
char *_NSPrintForDebugger (id obj)
{
    // Create a pool, call description, print out what it says, and delete the
    // pool. This will make sure any temporary objects that description might
    // create don't hang around.
    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
    printf ("%s\n" , [[obj description] cString]);
    [pool release];

    // Return a string with a space in it. gdb is smart enough to know 
    // that "" is no description, and advertises it as such.
    return " ";
}

/********************************************************************/

// Actually log a message. We reference the _FILE_, _FUNCTION_ and
// _LINE_ from the C version of the log routines.
static void logAMessage (FILE *out , const char *type , NSString *str)
{
    static const char *lastFile;
    static const char *lastFunction;
    static INT4 lastLine;
    int printed;

    printed = fprintf (out , "%s: " , type);
    printed += fprintf (out , "%s" , [str cString]);

    if (lastFile != _FILE_ || lastFunction != _FUNCTION_ || lastLine != _LINE_)
    {
        // If we printed out a small amount of information, then spit out
        // some seperating space between the message and the location,
        // so things don't look so crowded.
        if (printed < 80)
        {
            while (printed++ < 80)
                fprintf (out , " ");
        }

        // Something changed, so output it
        fprintf (out , " (");

        // If the file is different, output it.
        if (lastFile != _FILE_)
            fprintf (out , "%s" , _FILE_);
        else
            fprintf (out , "^");

        // if the line is different, output it
        if (lastLine != _LINE_)
            fprintf (out , ":%d" , _LINE_);

        // if the function is different, output it
        if (lastFunction != _FUNCTION_)
            fprintf (out , " in %s" , _FUNCTION_);

        fprintf (out , ")\n");
        
        // Save our state now
        lastFile = _FILE_;
        lastFunction = _FUNCTION_;
        lastLine = _LINE_;
    } else {
        fprintf (out , "\n");
    }
}

/********************************************************************/

// Logging functions
void MessageLog_ObjC_ (NSString *fmt , ...)
{
    va_list args;

    va_start (args , fmt);
    logAMessage (stdout , "Msg " , [NSString stringWithFormat: fmt arguments: args]);
    va_end (args);
}

/********************************************************************/

void WarningLog_ObjC_ (NSString *fmt , ...)
{
    va_list args;

    va_start (args , fmt);
    logAMessage (stdout , "Warn" , [NSString stringWithFormat: fmt arguments: args]);
    va_end (args);
}

/********************************************************************/

void ErrorLog_ObjC_ (NSString *fmt , ...)
{
    va_list args;

    va_start (args , fmt);
    logAMessage (stderr , "Err " , [NSString stringWithFormat: fmt arguments: args]);
    va_end (args);
}

/********************************************************************/

void FatalErrorLog_ObjC_ (NSString *fmt , ...)
{
    va_list args;

    va_start (args , fmt);
    logAMessage (stderr , "FErr" , [NSString stringWithFormat: fmt arguments: args]);
    va_end (args);
}

/********************************************************************/

void ProgrammerErrorLog_ObjC_ (NSString *fmt , ...)
{
    va_list args;

    va_start (args , fmt);
    logAMessage (stderr , "PErr" , [NSString stringWithFormat: fmt arguments: args]);
    va_end (args);
}

/********************************************************************/


