/*****************************************************************************/

#include <Gen/Gen.h>
#include <lua.h>
#include <lualib.h>
#include <lauxlib.h>
#include <query/query.h>
#include "cookie.h"

/*****************************************************************************/

/* This gets the raw cookie string for parsing. It's stored in an environment
 * variable that we want to fiddle with as we parse it. So, it first checks
 * to see if it can find the string. If it can, it allocates a new string, 
 * copies the data over, and then returns the copy. Otherwise, NULL is returned.
 *
 * The caller should free the memory this returns when it's done. */
static char *getCookieText (void)
{
    char *string;
    char *returnValue;

    /* First thing we're going to do is attempt to get the cookie string
     * from the environment, if possible. */
    string = getenv ("HTTP_COOKIE");
    if (string == NULL)
        return NULL;

    /* Duplicate the value of the variable and return that. */
    returnValue = (char *) malloc (strlen (string) + 1);
    strcpy (returnValue , string);
    return returnValue;
}

/*****************************************************************************/

/* Parse a string containing cookie values. The value of the string is
 * something like:
 *
 * Name=Value; Name=Value
 *
 * Where each name/value pair is separated by a semicolon if it exists. 
 *
 * This assumes that on entry the top of the stack is a table. This will
 * insert all the values into the table using the name as a key. 
 *
 * Parsing stops at the first invalid cookie string, or end of the string, 
 * whichever comes first.
 *
 * Note also that the cookie string passed in is modified during parsing. */
void parseCookieString (lua_State *L , char *cookieStr)
{
    char *key,*value;

    /* Start by trying to allocate memory for our buffers. If either of these
     * don't work, we should just return out right now. Notice that on  a failure 
     * of the second allocation, we're careful not to leak the first one. */
    key = (char *) malloc (MAX_KEY_LEN + 1);
    if (key == NULL)
        return;
    
    value = (char *) malloc (MAX_VALUE_LEN + 1);
    if (value == NULL)
    {
        free (key);
        return;
    }

    /* Loop until we hit the end of the string */
    while (*cookieStr != '\0')
    {
        /* Fish out the key string. The return value is the postion
         * of what made us stop, either the = sign, or a null byte.
         * If we hit a null byte, this isn't a valid key, so we can
         * skip out right now. */
        cookieStr = copyStr (cookieStr , key , MAX_KEY_LEN , '=');
        if (*cookieStr == '\0') break;

        /* Do the same thing for the value. The difference here is that
         * we don't mind if we hit the null byte, that just tells us
         * there are no more key/value pairs left. Which means, we
         * examine the return value. If it's not a null byte, then we
         * assume there's more data and increment the cmdStr to skip
         * past the ';' that splits keypairs and get past whitespace to
         * the start of the next key, so that the next iteration starts
         * at the next key. */
        cookieStr = copyStr (cookieStr + 1 , value , MAX_VALUE_LEN , ';');

        if (*cookieStr != '\0') 
        {
            /* Skip past the semicolon terminator that we stopped on. */
            cookieStr++;

            /* Skip all whitespace. */
            while (isspace (*cookieStr))
                cookieStr++;
        }

        /* Push the key, then the value, then set the table. */
        lua_pushstring (L , key);
        lua_pushstring (L , value);
        lua_settable (L , -3);
    }
 
    /* Blow away the buffers, we're done now. */
    free (key);
    free (value);
}

/*****************************************************************************/

/* Open up the cookie library. */
LUALIB_API void lua_cookielibopen (lua_State *L)
{
    char *cookieStr;

    /* First thing to do is put a new empty table onto the stack. */
    lua_newtable (L);

    /* Now we want to parse up the cookie string. First step, see if we can
     * even get one. */
    cookieStr = getCookieText ();
    if (cookieStr != NULL)
    {
        /* We got a cookie string, so parse it. It will put subtables into the
         * table we added to the top of the stack if there are any cookies.
         * If there are not, the table will end up empty, but that's ok.
         * Either way, release the cookie string once we're done with it. */
        parseCookieString (L , cookieStr);
        free (cookieStr);
    }

    /* Now that we're all done, we need to set a global variable using
     * this table. This is, fortunately, easy. */
    lua_setglobal (L , "cookies");
}

/*****************************************************************************/

