/*****************************************************************************/

#define __USE_BSD
#include <Gen/Gen.h>
#include <dirent.h>
#include <grp.h>
#include <pwd.h>
#include <libgen.h>
#include <lua.h>
#include <lualib.h>
#include <lauxlib.h>

/*****************************************************************************/

/* This defines the initial size of the encode buffer for the encoding functions,
 * as well as how much that buffer grows when it runs low on space. */
#define ENCODE_BUFFER_DELTA 1024

/* Assume that the current storage size of the encode buffer is size, and that
 * you want to add addition bytes to it. If the buffer is not big enough for that
 * to happen safely, then expand the buffer by ENCODE_BUFFER_DELTA to make room. 
 *
 * The buffer is expanded via realloc, so it's contents will not change. However,
 * you should not rely on the address of encodeBuffer staying the same (i.e. you
 * should access it via index, not via pointer math. */
#define SIZE_ENCODE_BUFFER(size,addition) if (size + addition >= encodeBufferSize)\
    { encodeBufferSize += ENCODE_BUFFER_DELTA;\
      encodeBuffer = realloc (encodeBuffer , encodeBufferSize); }

/*****************************************************************************/

/* These are used by the url encode/decode functions to hold a copy of the
 * string as it is being encoded/decoded before return. Note that this is not
 * threadsafe, but we're not really going to multithread a CGI script anyway. */
static char *encodeBuffer;
static INT4 encodeBufferSize;

/*****************************************************************************/

/* This is stored in crypt.c instead of here because we need to include unistd.h
 * with different flags. */
int ext_crypt (lua_State *L);

/*****************************************************************************/

/* A simple helper, it assumes the top of the stack is a table, and adds the
 * given number to the table with the given key name. */
static void helper_addnumber (lua_State *L , long value , char *keyName)
{
    /* Push the key and value, and then leave. */
    lua_pushstring (L , keyName);
    lua_pushnumber (L , value);
    lua_settable (L , -3);
}

/*****************************************************************************/

/* This is a helper for ext_stat. It takes a time value and a key name,
 * assumes that the top of the stack is a table, and inserts into that
 * table a table with the given key. The inserted table is the date, broken
 * apart into constituent parts. */
static void helper_stattime (lua_State *L , time_t timeVal , char *keyName)
{
    /* Insert the key. */
    lua_pushstring (L , keyName);

    /* Call the date function, telling it to return a table for the date
     * that we were given. */
    lua_getglobal (L , "date");
    lua_pushstring (L , "*t");
    lua_pushnumber (L , timeVal);
    lua_call (L , 2 , 1);

    /* Now we can set this key into the table. */
    lua_settable (L , -3);
}

/*****************************************************************************/

/* This is a helper for ext_stat. It takes a group ID and a key, and sets
 * into the table at the top of the stack a key/value pair that is the name
 * of that group. */
static void helper_groupid (lua_State *L , gid_t gid , char *keyName)
{
    struct group *grInfo;

    /* Push the key. */
    lua_pushstring (L , keyName);

    /* Look up the group name, if we can. If it works, push the group
     * name. If it doesn't, push the group number instead. */
    grInfo = getgrgid (gid);
    if (grInfo == NULL)
        lua_pushnumber (L , gid);
    else
        lua_pushstring (L , grInfo->gr_name);

    /* Go */
    lua_settable (L , -3);
}

/*****************************************************************************/

/* This is a helper for ext_stat. It sets the given key with the type of
 * the file, as given by the passed in mode paramter. */
static void helper_filetype (lua_State *L , mode_t mode , char *keyName)
{
    /* Push the key */
    lua_pushstring (L , keyName);

    /* Push the type */
    if (S_ISREG  (mode)) lua_pushstring (L , "file"); else
    if (S_ISDIR  (mode)) lua_pushstring (L , "directory"); else
    if (S_ISCHR  (mode)) lua_pushstring (L , "charDevice"); else
    if (S_ISBLK  (mode)) lua_pushstring (L , "blockDevice"); else
    if (S_ISFIFO (mode)) lua_pushstring (L , "fifo"); else
    if (S_ISLNK  (mode)) lua_pushstring (L , "link"); else
    if (S_ISSOCK (mode)) lua_pushstring (L , "socket");  else
        lua_pushstring (L , "unknown");

    /* Go */
    lua_settable (L , -3);
}

/*****************************************************************************/

/* This is a helper for ext_stat. It takes a user ID and a key, and sets
 * into the table at the top of the stack a key/value pair that is the name
 * of that user. */
static void helper_userid (lua_State *L , uid_t uid , char *keyName)
{
    struct passwd *usInfo;

    /* Push the key. */
    lua_pushstring (L , keyName);

    /* Look up the user name, if we can. If it works, push the user
     * name. If it doesn't, push the user number instead. */
    usInfo = getpwuid (uid);
    if (usInfo == NULL)
        lua_pushnumber (L , uid);
    else
        lua_pushstring (L , usInfo->pw_name);

    /* Go */
    lua_settable (L , -3);
}

/*****************************************************************************/

/* This is a helper for ext_stat. It takes 3 bools that indicate read/write/exec
 * permissions, and then creates a table with the given keyname that has keys
 * for those three things. This table is set into the table on the top of the
 * stack. */
static void helper_permissions (lua_State *L , int r , int w , int x , char *keyName)
{
    /* Push the key, then add a new table. */
    lua_pushstring (L , keyName);
    lua_newtable (L);

    /* For each permission, add an entry to the table, if it is set. */
    if (r) helper_addnumber (L , 1 , "read");
    if (w) helper_addnumber (L , 1 , "write");
    if (x) helper_addnumber (L , 1 , "execute");

    /* Set it in */
    lua_settable (L , -3);
}

/*****************************************************************************/

/* dircontents ([dirname])
 *
 * Returns to you a table contains the contents of the passed in directory, or 
 * the contents of the current directory, if no directory name is given. If 
 * there is an error, nil and an error message are returned instead. 
 *
 * This will not include the . and .. directories, since these are assumed to
 * be everywhere. */
static int ext_dircontents (lua_State *L) 
{
    const char *startDir = luaL_opt_lstr (L , 1 , "." , NULL);
    struct dirent **nameList;
    int retVal,i,added=0;

    /* Scan the directory given to attempt to get the contents */
    retVal = scandir (startDir , &nameList , NULL , alphasort);

    /* If the return value is -1, it means that something has gone
     * horribly wrong. So, return back nil and an error message. */
    if (retVal == -1)
    {
        lua_pushnil (L);
        lua_pushstring (L , strerror (errno));
        return 2;
    }

    /* We got something. It might be nothing, though. Put a table on the stack,
     * and then insert all the items. */
    lua_newtable (L);

    for (i = 0 ; i < retVal ; i++)
    {
        /* If the name is "." or "..", then skip it. */
        if (strcmp (nameList[i]->d_name , ".") == 0 || strcmp (nameList[i]->d_name , "..") == 0)
        {
            free (nameList[i]);
            continue;
        }

        /* Push the name of this item onto the stack, then free it. Finally,
         * set it into the table at the current index. */
        lua_pushstring (L , nameList[i]->d_name);
        free (nameList[i]);
        lua_rawseti (L , -2 , added + 1);
        added++;
    }

    /* Now that we're done, throw the name list away. */
    free (nameList);

    /* For a bit of a speedup, insert the total amount of entries as "n".
     * the function getn () checks for this value before counting. */
    lua_pushstring (L , "n");
    lua_pushnumber (L , added);
    lua_settable (L , -3);

    /* Return the table back */
    return 1;
}

/*****************************************************************************/

/* stat (filename)
 *
 * Returns back a table that contains information about the given file, or
 * nil and an error message, if something goes horribly wrong. */
static int ext_stat (lua_State *L)
{
    /* Get the filename that we will stat */
    const char *fname = luaL_check_lstr (L , 1 , NULL);
    int doLinkStat;
    int result;
    struct stat statBuf;

    /* Determine if we will lstat or not. The second paramter has to be
     * non-nil to lstat. */
    if (lua_isnoneornil (L , 2))
        doLinkStat = 0;
    else
        doLinkStat = 1;

    /* Attempt the stat. If it doesn't work, then bail out. */
    if (doLinkStat)
        result = lstat (fname , &statBuf);
    else
        result = stat (fname , &statBuf);

    if (result != 0)
    {
        lua_pushnil (L);
        lua_pushstring (L , strerror (errno));
        return 2;
    } 

    /* Must have worked. Put a table on the stack so we can set stuff
     * into it. */
    lua_newtable (L);

    /* Insert the stat type that we did. This is primarily for debugging, but
     * someone might want to know after the fact the type, since with a
     * non-lstat, you can never tell if anything is a link. */
    lua_pushstring (L , "statType");
    if (doLinkStat)
        lua_pushstring (L , "lstat");
    else
        lua_pushstring (L , "stat");
    lua_settable (L , -3);

    /* Insert the size of the file */
    helper_addnumber (L , statBuf.st_size , "size");

    /* Insert the user and group ID */
    helper_addnumber (L , statBuf.st_uid , "ownerUID");
    helper_userid    (L , statBuf.st_uid , "user");
    helper_addnumber (L , statBuf.st_gid , "ownerGID");
    helper_groupid   (L , statBuf.st_gid , "group");
   
    /* Insert the access time */
    helper_stattime (L , statBuf.st_atime , "accessTime");

    /* Insert the modification time */
    helper_stattime (L , statBuf.st_mtime , "modificationTime");

    /* Insert the change time */
    helper_stattime (L , statBuf.st_ctime , "changeTime");

    /* Insert the file type */
    helper_filetype (L , statBuf.st_mode , "type");

    /* Insert the owner permissions. */
    helper_permissions (L , statBuf.st_mode & S_IRUSR ,
                            statBuf.st_mode & S_IWUSR , 
                            statBuf.st_mode & S_IXUSR , "ownerPermissions");

    /* Insert the group permissions. */
    helper_permissions (L , statBuf.st_mode & S_IRGRP ,
                            statBuf.st_mode & S_IWGRP , 
                            statBuf.st_mode & S_IXGRP , "groupPermissions");
    
    /* Insert the owner permissions. */
    helper_permissions (L , statBuf.st_mode & S_IROTH ,
                            statBuf.st_mode & S_IWOTH , 
                            statBuf.st_mode & S_IXOTH , "otherPermissions");

    /* If the setUID bit is set, then add that in. Do the same for setGID
     * and sticky. */
    if (statBuf.st_mode & S_ISUID) helper_addnumber (L , 1 , "setUID");
    if (statBuf.st_mode & S_ISGID) helper_addnumber (L , 1 , "setGID");
    if (statBuf.st_mode & S_ISVTX) helper_addnumber (L , 1 , "isSticky");

    /* Return the table back */
    return 1;
}

/*****************************************************************************/

/* readlink (filename)
 *
 * Assumes that the passed in filename is a symlink. This will query to see
 * where the link is pointing. That is returned. Otherwise, nil and an error
 * message is returned. */
static int ext_readlink (lua_State *L)
{
    /* Get the filename that we will check */
    const char *fname = luaL_check_lstr (L , 1 , NULL);
    char fbuffer[256];
    int result;

    /* Do the call, check for error. */
    if ((result = readlink (fname , fbuffer , sizeof (fbuffer) - 1)) == -1)
    {
        lua_pushnil (L);
        lua_pushstring (L , strerror (errno));
        return 2;
    }

    /* Put the string. Since it's not null terminated, tell the other side how
     * big it is. */
    lua_pushlstring (L , fbuffer , result);
    return 1;
}

/*****************************************************************************/

/* dirname (filename)
 * 
 * Returns the path portion of the given filename. */
static int ext_dirname (lua_State *L)
{
    /* Get the filename that we will check */
    const char *fname = luaL_check_lstr (L , 1 , NULL);
    char buffer[256];

    /* Make a copy */
    strncpy (buffer , fname , sizeof (buffer) - 1);
    buffer[sizeof (buffer) - 1] = '\0';

    /* Do the job and return the string. */
    lua_pushstring (L , dirname (buffer));
    return 1;
}

/*****************************************************************************/

/* basename (filename)
 *
 * Returns the filename portion of the given filename. */
static int ext_basename (lua_State *L)
{
    /* Get the filename that we will check */
    const char *fname = luaL_check_lstr (L , 1 , NULL);
    char buffer[256];

    /* Make a copy */
    strncpy (buffer , fname , sizeof (buffer) - 1);
    buffer[sizeof (buffer) - 1] = '\0';

    /* Do the job and return the string. */
    lua_pushstring (L , basename (buffer));
    return 1;
}

/*****************************************************************************/

/* This helper function tells you if the character passed in needs to be
 * URL encoded. */
static BOOL helper_charNeedsEncoding (char c)
{
    /* Spaces and punctuation (i.e. non alphanumeric) characters need to be
     * encoded. */
    if (isspace (c) || ispunct (c))
        return YES;

    /* Must be ok. */
    return NO;
}

/*****************************************************************************/

/* urlencode (string)
 *
 * Return a copy of the passed in string which has been URL encoded (all 
 * non-alphanumeric characters converted to %XX, where XX is the hex value of
 * the character. */
static int extstr_urlencode (lua_State *L)
{
    /* Get the string to work on. */
    const char *inStr = luaL_check_lstr (L , 1 , NULL);
    INT4 srcIndex=0,dstIndex=0;
    char hexBuf[3];

    /* Keep encoding the input string until we finish it. */
    while (inStr[srcIndex] != '\0')
    {
        /* If the character at the source index doesn't need encoding, then
         * we are only copying 1 character. Otherwise, we're copying 3
         * characters (%xx). */
        if (helper_charNeedsEncoding (inStr[srcIndex]) == NO)
        {
            /* Make sure the encode buffer is big enough, then copy it over. */
            SIZE_ENCODE_BUFFER (dstIndex, 1);
            encodeBuffer[dstIndex++] = inStr[srcIndex++];
        } 
        else
        {
            /* Same as above, but more characters. We also need to set up the
             * hex buffer we're copying over. */
            SIZE_ENCODE_BUFFER (dstIndex, 3);

            /* Convert the character at the source index to it's hex equivilant,
             * and skip past the character. */
            sprintf (hexBuf , "%0X" , inStr[srcIndex]);
            srcIndex++;

            /* Now copy the encoded character over. */
            encodeBuffer[dstIndex++] = '%';
            encodeBuffer[dstIndex++] = hexBuf[0];
            encodeBuffer[dstIndex++] = hexBuf[1];
        }
    }

    /* Return the output string. We will use pushlstring here so that we can
     * tell lua the length. That way we don't have to null terminate it (and
     * maybe expand the buffer again as a result. */
    lua_pushlstring (L , encodeBuffer , dstIndex);
    return 1;
}

/*****************************************************************************/

/* urldecode (string)
 *
 * Reverses urlencode. All url like sequences in the string will be undone
 * and reverted back to the normal character. */
static int extstr_urldecode (lua_State *L)
{
    /* Get the string to work on. */
    const char *inStr = luaL_check_lstr (L , 1 , NULL);
    INT4 srcIndex=0,dstIndex=0;
    INT4 hexVal;
    char hexBuf[3] = {0 , 0 , 0};

    /* Decode the entire input string. */
    while (inStr[srcIndex] != '\0')
    {
        /* We are going to be copying a character over, so make sure that the
         * encode buffer will be big enough to handle it. */
        SIZE_ENCODE_BUFFER (dstIndex , 1);

        /* If the character at srcIndex is not a '%' charcter, then it is safe
         * to just copy it over. */
        if (inStr[srcIndex] != '%')
        {
            /* Copy the character over as is. */
            encodeBuffer[dstIndex++] = inStr[srcIndex++];
        }
        else
        {
            /* This character is a '%' character. Skip it, and save the next two characters,
             * which are hex digits. */
            srcIndex++;
            hexBuf[0] = inStr[srcIndex++];
            hexBuf[1] = inStr[srcIndex++];

            /* Convert the hex string back into a character, then add it to the 
             * decoded buffer. */
            sscanf (hexBuf , "%x" , &hexVal);
            encodeBuffer[dstIndex++] = hexVal;
        }
    }

    /* Return the output string. We will use pushlstring here so that we can
     * tell lua the length. That way we don't have to null terminate it (and
     * maybe expand the buffer again as a result. */
    lua_pushlstring (L , encodeBuffer , dstIndex);
    return 1;
}

/*****************************************************************************/

/* trimleading (string)
 *
 * Trims leading whitespace off of the given string.
 */
static int extstr_trimleading (lua_State *L)
{
    /* Get the string to work on. */
    const char *inStr = luaL_check_lstr (L , 1 , NULL);

    /* While the first character is whitespace, skip it. */
    while (isspace (*inStr))
        inStr++;

    /* Now push the new string back. */
    lua_pushstring (L, inStr);
    return 1;
}

/*****************************************************************************/

/* trimtrailing (string)
 *
 * Trims trailing whitespace off of the given string.
 */
static int extstr_trimtrailing (lua_State *L)
{
    /* Get the string to work on. */
    const char *inStr = luaL_check_lstr (L , 1 , NULL);
    int length = strlen (inStr);

    /* While the character at length - 1 is whitespace, decrement
     * length by 1. */
    while (length > 0 && isspace (inStr[length - 1]))
        length--;

    /* Now push the new string back. */
    lua_pushlstring (L, inStr, length);
    return 1;
}

/*****************************************************************************/

/* trimleadingtrailing (string)
 *
 * Trims leading and trailing whitespace off of the given string.
 */
static int extstr_trimleadingtrailing (lua_State *L)
{
    /* Get the string to work on. */
    const char *inStr = luaL_check_lstr (L , 1 , NULL);
    int length;

    /* While the first character is whitespace, skip it. */
    while (isspace (*inStr))
        inStr++;

    /* Get length */
    length = strlen (inStr);

    /* While the character at length - 1 is whitespace, decrement
     * length by 1. */
    while (length > 0 && isspace (inStr[length - 1]))
        length--;

    /* Now push the new string back. */
    lua_pushlstring (L, inStr , length);
    return 1;
}

/*****************************************************************************/

/* strcmp (string, string)
 *
 * Works like C strcmp does.
 */
static int extstr_strcmp (lua_State *L)
{
    /* Get the strings to work on. */
    const char *first  = luaL_check_lstr (L , 1 , NULL);
    const char *second = luaL_check_lstr (L , 2 , NULL);
    int result;

    /* Do the comparison and push the result. */
    result = strcmp (first , second);
    lua_pushnumber (L, result);
    return 1;
}

/*****************************************************************************/

/* strcasecmp (string, string)
 *
 * Works like strcasecmp does.
 */
static int extstr_strcasecmp (lua_State *L)
{
    /* Get the strings to work on. */
    const char *first  = luaL_check_lstr (L , 1 , NULL);
    const char *second = luaL_check_lstr (L , 2 , NULL);
    int result;

    /* Do the comparison and push the result. */
    result = strcasecmp (first , second);
    lua_pushnumber (L, result);
    return 1;
}

/*****************************************************************************/

/* mktime (year, month, day, hour, min)
 *
 * Take a date and convert it to an epoch time. This works similarly to
 * os.time would to create a time from other information, except that the
 * core lua version needs to know if the date is dst or not.
 *
 * This version always creates the date in gmt. */
static int extos_mktime (lua_State *L)
{
    struct tm ts;
    time_t result;
    int year   = luaL_check_number (L, 1);
    int month  = luaL_check_number (L, 2);
    int day    = luaL_check_number (L, 3);
    int hour   = luaL_check_number (L, 4);
    int minute = luaL_check_number (L, 5);

    /* Copy the values over, fixing up the year and month as we go. */
    ts.tm_year = year - 1900;
    ts.tm_mon = month - 1;
    ts.tm_mday = day;
    ts.tm_hour = hour;
    ts.tm_min = minute;
    ts.tm_sec = 0;

    /* The -1 tells the system that it's unknown if this date
     * should be daylight savings time or not. */
    ts.tm_isdst = -1;

    /* Now convert the value and push it. */
    result = mktime (&ts);

    if (result == -1)
        lua_pushnil (L);
    else
        lua_pushnumber (L, result);
    
    /* One value returned. */
    return 1;
}

/*****************************************************************************/

static const struct luaL_reg extendedlib[] = {
  {"dircontents",     ext_dircontents},
  {"stat",            ext_stat},
  {"readlink",        ext_readlink},
  {"dirname",         ext_dirname},
  {"basename",        ext_basename},
  {"crypt",           ext_crypt},
  {NULL,              NULL},
};

/*****************************************************************************/

static const struct luaL_reg extendedstringlib[] =
{
    {"urlencode"           , extstr_urlencode},
    {"urldecode"           , extstr_urldecode},
    {"trimleading"         , extstr_trimleading},
    {"trimtrailing"        , extstr_trimtrailing},
    {"trimleadingtrailing" , extstr_trimleadingtrailing},
    {"strcmp"              , extstr_strcmp},
    {"strcasecmp"          , extstr_strcasecmp},
    {NULL, NULL},
};

/*****************************************************************************/

static const struct luaL_reg extendedoslib[] =
{
    {"mktime"           , extos_mktime},
    {NULL, NULL},
};

/*****************************************************************************/

/* Open up the extensions library. */
LUALIB_API void lua_extensionslibopen (lua_State *L)
{
    /* Seed the random number generator with the current time, so that every run
     * we can get different random values. */
    srand (time (NULL));

    /* Set up the initial encode buffer to be 1k in size. This will be expanded
     * as needed at runtime. */
    encodeBufferSize = ENCODE_BUFFER_DELTA;
    encodeBuffer = (char *) malloc (encodeBufferSize * sizeof (char));

    /* Add our extension functions */
    luaL_openlib (L, "ext" , extendedlib , 0);
    luaL_openlib (L , "string" , extendedstringlib , 0);
    luaL_openlib (L , "os" , extendedoslib , 0);

}

/*****************************************************************************/

