/*****************************************************************************/

#include <Gen/Gen.h>
#include <lua.h>
#include <lualib.h>
#include <lauxlib.h>
#include "graphics.h"

/*****************************************************************************/

/* These are used in the code that gets the size of a jpeg. */
#define readbyte(a,b) do if (((a) = getc((b))) == EOF) { fclose (b); return 0; } while (0)
#define readword(a,b) do { int cc_=0,dd_=0; \
                           if((cc_=getc((b))) == EOF || (dd_=getc((b))) == EOF) { fclose (b); return 0; } \
                           (a) = (cc_<<8) + (dd_); \
                         } while(0)

/*****************************************************************************/

/* Checks the size of a jpeg file, returning the dimensions in the values
 * given. If it works, 1 is returned, else 0 is returned. */
static int jpeg_size (char *filename, int *image_width , int *image_height) 
{
    int marker=0;
    int dummy=0;
    FILE *infile;

    infile = fopen (filename , "rb");
    if (infile == NULL)
        return 0;
  
    /* Eat the file header. Leave if we don't get the bytes we think we
     * should. */
    if (getc (infile) != 0xFF || getc (infile) != 0xD8)
    {
        fclose (infile);
        return 0;
    }

    /* Loop forever reading bytes, because jpeg files store data in chunks,
     * like wave files, and the chunk with the dimensions aren't neccesarily
     * at the beginning. */
    while (1)
    {
        int discarded_bytes=0;
        
        /* Read a marker. */
        readbyte (marker,infile);

        /* If this marker isn't 0xFF, then throw away bytes until we find
         * one that is. */
        while (marker != 0xFF) 
        {
            discarded_bytes++;
            readbyte (marker,infile);
        }
        
        /* Now, read bytes until we hit a 0xFF again. */
        do 
            readbyte(marker,infile); 
        while (marker == 0xFF);

        /* If we had to discard any bytes to do that, then this isn't a valid
         * jpeg file. */
        if (discarded_bytes != 0) 
        {
            fclose (infile);
            return 0;
        }
   
        /* Check the marker to see what's what. */
        switch (marker) 
        {
            /* Chunk types that carry dimension information. */
            case 0xC0: case 0xC1: case 0xC2: case 0xC3: case 0xC5: case 0xC6:
            case 0xC7: case 0xC9: case 0xCA: case 0xCB: case 0xCD: case 0xCE:
            case 0xCF: 
            {
                /* Get chumk parameter length, a filler byte, then we can
                 * get the dimensions out, and return. */
                readword (dummy , infile);    
                readbyte (dummy , infile);
                readword ((*image_height) , infile);
                readword ((*image_width) , infile);
                fclose (infile);

                return 1;
            }

            /* If we hit these, that's it for us. I'm not sure, but I think that
             * they indicate the end of the file, and so if we've hit them,
             * we already know that we can't get any info out. */
            case 0xDA:
            case 0xD9:
                fclose (infile);
                return 0;

            /* All other chunk types. Just read and discard. */
            default: 
            {
                int length;
        
                /* Get chunk length. It includes itself, so it's always two
                 * bigger than you'd think. */
                readword (length,infile);

                /* Chunk has no length? Then leave. */
                if (length < 2)
                {
                    fclose (infile);
                    return 0;
                }

                /* Take the length of the length into account, then throw the
                 * rest of the bytes away. */
                length -= 2;
                while (length > 0) 
                {
                    readbyte (dummy, infile);
                    length--;
                }
            }
            break;
        }
    }
}

/*****************************************************************************/

/* A simple lua function that can tell you the size of a jpeg file. 
 *
 * jpegsize (name) - Return back width,height of the image, or nil,nil if the
 *                   file is missing/not a jpeg. */
static int graphics_jpegsize (lua_State *L)
{
    int width,height;
    char *fname;

    /* Get the filename or error out. */
    fname = (char *) luaL_check_string (L , 1);

    /* Try to determine the size of that file. */
    if (jpeg_size (fname , &width , &height) == 0)
        return 0;

    /* Push the results */
    lua_pushnumber (L , width);
    lua_pushnumber (L , height);

    /* Returning 2 things. */
    return 2;
}

/*****************************************************************************/

static const struct luaL_reg graphicslib[] = {
    {"jpeg_size" , graphics_jpegsize},
    {NULL        , NULL},
};

/*****************************************************************************/

/* Open up the graphics library. */
LUALIB_API void lua_graphicslibopen (lua_State *L)
{
    /* Add our graphics functions. */
    luaL_openlib (L , "graphics" , graphicslib , 0);
}

/*****************************************************************************/

