-- This file implements a simple HTML generation library in lua. It allows you to add cookie
-- and textual content in various ways, and even change the content type. When you call the
-- page generation function, all content up to this point will be flushed out, including the
-- proper headers to tell the browser what we're dumping out and any cookies.
--
-- Once a page has been generated in that way, it will be "flushed" out, so that subsequent
-- calls to the generation function will only write out things added since the last time the
-- page was written out. This allows for incremental page building.
--
-- Because the content type and cookies must be the first things out, once the page has been
-- generated once, you cannot add any more cookies or change the content type. Doing so will
-- be ignored.
html = {
  ["__content_type"] = "text/html",
  ["__title"] = "untitled page" ,
  ["__styles"] = nil,
  ["__bgcolor"] = nil,
  ["__background"] = nil,
  ["__text"] = nil,
  ["__link"] = nil,
  ["__alink"] = nil,
  ["__vlink"] = nil,
  ["__flushed"] = false,
  ["__lists"] = {},
  ["__cookies"] = {},
  ["__content"] = {}
}

html.table = {
  ["__tables"] = {},
}

html.form = {
  ["__formopen"] = "no"
}

-- This will make handling forms easier. It will make the formData element of CGI be either
-- the parsedMultipartData or parsedQueryString or an empty table, depending on which form 
-- method was used.
CGI.formData = CGI.parsedMultipartData or CGI.parsedQueryString or {}

-------------------------------------------------------------------------------------------

-- Output an HTML attribute. If value is nil, does nothing. Otherwise, adds an attribute
-- with the given name and value to the current tag, by outputting the attribute preceeded
-- by whitespace.
local attribute = function (name, value)
    if value then
        html:addText (' ' , name , '="' , value , '"')
    end
end

-------------------------------------------------------------------------------------------

-- Tells you if the page has been flushed or not.
function html:wasFlushed ()
    return self.__flushed
end

-------------------------------------------------------------------------------------------

-- Set the content type of the page. 
function html:contentType (type)
    -- Can't do this if the page was flushed.
    if html:wasFlushed () then return end
    self.__content_type = type
end

-------------------------------------------------------------------------------------------

-- Set the title of the page being generated. This must be done prior to the page being
-- flushed for the first time or it will have no effect.
function html:setTitle (title)
    -- Can't do this if the page was flushed.
    if html:wasFlushed () then return end
    self.__title = title
end

-------------------------------------------------------------------------------------------

-- Get the current page title, as it will be written out, or as it was written out if
-- the page was already flushed.
function html:getTitle ()
    return self.__title
end

-------------------------------------------------------------------------------------------

-- Set colors for the page. Each argument becomes a part of the generated body tag in the
-- page. This must be called before the page is flushed for the first time or it will have
-- no effect.
--
-- The function ignores arguments passed in as nil, so you can use this to selectively change
-- items and leave others alone. To disable a setting, pass in an empty string for it.
function html:setColors (background,bgcolor,text,link,alink,vlink)
    -- Can't do this if the page was flushed.
    if html:wasFlushed () then return end

    if background  then self.__background = background end
    if bgcolor     then self.__bgcolor    = bgcolor end
    if text        then self.__text       = text end
    if link        then self.__link       = link end
    if alink       then self.__alink      = alink end
    if vlink       then self.__vlink      = vlink end
end

-------------------------------------------------------------------------------------------

-- This function can be used to link to a style sheet for the generated page, provided that
-- it is called before any page data is output. href is a url to the stylesheet to load.
-- If nil, it defaults to "/stylesheet.css". styleType defines the type of the stylesheet.
-- If not given, it defaults to "text/css".
--
-- This will put a link section into the head of the page when it is generated.
function html:setStylesheet (href,styleType)
    -- Can't do this if the page was flushed.
    if html:wasFlushed () then return end

    -- Make both values default if not specified
    styleType = styleType or "text/css"
    href      = href      or "/stylesheet.css"

    -- Add a styles memeber to the structure and fill it in
    self.__styles = {}
    self.__styles.styleType = styleType
    self.__styles.href = href
end

-------------------------------------------------------------------------------------------

-- Add a cookie of the given name and value. Only the name and value parts are required.
--
-- path is optional, and specifies what paths the cookie should be sent
-- back to on further requests. A path of "/" is the most general path, and matches
-- any path. If not given, path is assumed to be the path of the document that is setting
-- the cookie, such that only files served from the same location as the path will get
-- the cookie back.
--
-- expireTime is optional. If it is set, it should be an integer that represents an epoch
-- time for when the cookie should expire. Note that not all browsers will honour this 
-- expiration time, so you need to be prepared to get a cookie even if it should have
-- expired. 
--
-- If no date is given, the cookie is only valid for the current browser session. If you
-- persist a cookie by giving a date, you can force the browser to discard the cookie
-- sooner by re-setting the same cookie with an expire time that is in the past. This
-- makes the browser drop the cookie right away (but see above).
function html:addCookie (name , value , expireTime , path)
    -- Can't do this if the page was flushed.
    if html:wasFlushed () then return end

    -- Create the cookie
    local newCookie = {}
    newCookie.name = string.urlencode (name)
    newCookie.value = string.urlencode (value)
    newCookie.path = path
    newCookie.expires = expireTime

    -- Add it
    table.insert (self.__cookies , newCookie)
end

-------------------------------------------------------------------------------------------

-- This function removes a cookie so that the browser will (probably) no longer send it.
-- The name and path passed in here must be identical to the name and path passed in when
-- the cookie was added for it to be removed.
function html:removeCookie (name , path)
    html:addCookie (name , "deleted" , 0 , path)
end

-------------------------------------------------------------------------------------------

-- Take a list of strings, and output them all to stderr, along with a trailing newline.
-- Used for debugging scripts. The output will get sent to the error log.
function html:debug (...)
    -- Loop over all arguments given
    for index,text in ipairs (arg) do
        -- If this is a table, handle it special. Otherwise,
        -- insert it as is.
        if type (text) == "table" then
            for inner,iText in ipairs (text) do
                io.stderr:write (tostring (iText))
            end
        else
            io.stderr:write (tostring (text))
        end
    end

    io.stderr:write ("\n")
end

-------------------------------------------------------------------------------------------

-- Add a string of any length to the page. This will be added
-- exactly as is. Any argument that is a table will be assumed to
-- be an array of strings, which will be added. Note that this
-- means that the table can't contain a table.
function html:addText (...)
    -- Loop over all arguments given
    for index,text in ipairs (arg) do
        -- If this is a table, handle it special. Otherwise,
        -- insert it as is.
        if type (text) == "table" then
            for inner,iText in ipairs (text) do
                table.insert (self.__content , tostring (iText))
            end
        else
            table.insert (self.__content , tostring (text))
        end
    end
end

-------------------------------------------------------------------------------------------

-- Use the current state of things to generate an HTML page
-- to stdout.
function html:flushPage ()
    -- If the page hasn't been flushed yet, then write out header
    -- information.
    if html:wasFlushed () == false then
        -- To start, write the content type.
        io.write ("Content-type: " ,  self.__content_type , "\n")

        -- Write all cookies.
        for index,cookie in ipairs (self.__cookies) do
            io.write ("Set-cookie: " , cookie.name , "=" , cookie.value , ";")
            if cookie.path then io.write (" path=" , cookie.path , ";") end
            if cookie.expires then io.write (" expires=" , os.date ("%a, %d-%b-%Y %H:%M:%S GMT" , cookie.expires) , ";") end
            io.write ("\n")
        end

        -- Terminate headers with a newline now.
        io.write ("\n")

        -- Now start the page. First, open the HTML tag and then set the page title
        io.write ("<HTML><HEAD><TITLE>" , self.__title , "</TITLE>")

        -- If there is style information attached, include it now.
        if self.__styles then
            io.write ('<LINK REL="STYLESHEET" TYPE="' , self.__styles.styleType , '" HREF="' , self.__styles.href , '" />')
        end 

        -- Done with the head, open the body tag now
        io.write ("</HEAD><BODY")

        -- For each body element, if it is set then put it into the page.
        if self.__background then io.write (' BACKGROUND="' , self.__background , '"') end
        if self.__bgcolor    then io.write (' BGCOLOR="'    , self.__bgcolor , '"') end
        if self.__text       then io.write (' TEXT="'       , self.__text , '"') end
        if self.__link       then io.write (' LINK="'       , self.__link , '"') end
        if self.__alink      then io.write (' ALINK="'      , self.__alink , '"') end
        if self.__vlink      then io.write (' VLINK="'      , self.__vlink , '"') end

        -- Close the body tag now.
        io.write (">\n")

        -- Now, mark the page as flushed.
        self.__flushed = true
    end

    -- Write out all page content in order.
    for index,text in ipairs (self.__content) do
        io.write (text)
    end

    -- Empty out the content table, since it has been written now
    self.__content = {}
end

-------------------------------------------------------------------------------------------

-- Use this to end a page. It will close the open body and html tags, then cause a flush.
function html:endPage ()
    -- While there are still lists open, close them up before we close the page.
    while  table.getn (self.__lists) ~= 0 do html:closeList () end

    -- Now do the same for tables (since the lists may have been inside a table)
    while  table.getn (self.table.__tables) ~= 0 do html.table:close () end

    -- Close any open form
    html.form:close ()

    html:addText ("</BODY></HTML>\n")
    html:flushPage ()
end

-------------------------------------------------------------------------------------------

-- Add a hyperlink to the page using the provided URL and link text.
function html:link (url , linkText , target, class, style, onclick)
    html:addText ('<A HREF="' , url , '"')

    attribute ("TARGET" , target)
    attribute ("CLASS" , class)
    attribute ("STYLE" , style)
    attribute ("ONCLICK", onclick)

    html:addText ('>' , linkText , "</A>")
end

-------------------------------------------------------------------------------------------

-- Add in a named spot to allow a hyperlink jump. This is like a regular
-- link tag except that it doesn't go anywhere.
function html:namedSpot (name)
    if name then
        html:addText ('<A')
        attribute ("NAME", name)
        html:addText ('></A>')
    end
end

-------------------------------------------------------------------------------------------

-- Add an image to the page using the provided properties. The most used properties are
-- included first. Any property that is set to nil will not be present. The only exception
-- is borderSize. If this is not given, it is assumed to be 0 (no border)
function html:image (src , width, height, alt , border, lowSrc, align, vspace, hspace)
    html:addText ("<IMG")

    attribute ("SRC" , src)
    attribute ("ALT" , alt)
    attribute ("WIDTH" , width)
    attribute ("HEIGHT" , height)
    attribute ("LOWSRC" , lowSrc)
    attribute ("ALIGN" , align)
    attribute ("VSPACE" , vspace)
    attribute ("HSPACE" , hspace)

    if border then attribute ("BORDER" , border) else attribute ("BORDER" , 0) end

    html:addText (">")
end

-------------------------------------------------------------------------------------------

-- A mix of the two functions above. Generate a link around an image
function html:linkImage (url , target , src , width , alt , height , border , lowSrc , align , vspace , hspace)
    html:addText ('<A HREF="' , url , '"')

    attribute ("TARGET" , target)
    html:addText ('>')

    html:image (src , width , height , alt , border , lowSrc , align , vspace , hspace)
    html:addText ('</A>')
end

-------------------------------------------------------------------------------------------

-- Output a header of the given type, wrapping all other arguments.
function html:header(size,...)
    html:addText ("<H" , size , ">")
    html:addText (unpack (arg))
    html:addText ("</H" , size , ">")
end

-------------------------------------------------------------------------------------------

-- Open an ordered list. list type can be one of "A" , "a" , "I" , "i" , "1" to set the
-- default numbering style of the list. startValue is the value to start the list at.
-- The default is to use "1" , "1", that is, a numbered list starting at 1.
function html:orderedList (listType,startValue,class,style)
    -- Open the tag
    html:addText ("<OL")

    attribute ("TYPE" , listType)
    attribute ("VALUE" , startValue)
    attribute ("CLASS" , class)
    attribute ("STYLE" , style)

    html:addText (">")

    -- Store this in the list stack.
    table.insert (self.__lists , "OL")
end

-------------------------------------------------------------------------------------------

-- Open an unordered list. list type can be one of "disc" , "circle" or "square".
function html:unorderedList (listType,class,style)
    -- Open the tag
    html:addText ("<UL")

    attribute ("TYPE" , listType)
    attribute ("CLASS" , class)
    attribute ("STYLE" , style)

    html:addText (">")

    -- Store this in the list stack.
    table.insert (self.__lists , "UL")
end

-------------------------------------------------------------------------------------------

-- Close the most recently opened list. This uses a stack to remember what the last list
-- opened was, and will generate the proper close tag to close that list.
function html:closeList ()
    -- Leave if someone tries to close a list when one isn't open.
    if  table.getn (self.__lists) == 0 then return end

    -- Get the last item from the table, then remove it.
    local listType = table.remove (self.__lists)

    html:addText ("</" , listType , ">")
end

-------------------------------------------------------------------------------------------

-- Takes all parameters and dumps them out following a list item tag.
function html:listItem (...)
    html:addText ("<LI>" , unpack (arg))
end

-------------------------------------------------------------------------------------------

-- As above, but pass in a style that is one of "A", "a" , "I" , "i" or "1" to set the
-- style of this particular item. Value can be nil, or the value to give this item (basically
-- to change the numeric sequence of this list item.
function html:styledListItem (style,value,...)
    html:addText ('<LI TYPE="' , style , '"')
    attribute ("VALUE" , value)
    html:addText (">" , unpack (arg))
end

-------------------------------------------------------------------------------------------

-- Open a new table. Each parameter works the way you would expect. Anything passed in as
-- nil will not be put into the table tag.
function html.table:open (border, width , height, cellSpacing, cellPadding, class, style, bgcolor, bordercolor)
    html:addText ("<TABLE")

    attribute ("BORDER" , border)
    attribute ("WIDTH" , width)
    attribute ("HEIGHT" , height)
    attribute ("CELLSPACING" , cellSpacing)
    attribute ("CELLPADDING" , cellPadding)
    attribute ("BGCOLOR" , bgcolor)
    attribute ("BORDERCOLOR" , bordercolor)
    attribute ("CLASS" , class)
    attribute ("STYLE" , style)

    html:addText (">\n")
    
    -- Store this in the table stack.
    table.insert (self.__tables , "empty")
end

-------------------------------------------------------------------------------------------

-- This private function will close any table subtag (TR, TH or TD) that is currently open.
-- If the current tag is TR, the row is closed and the table tag is set to "empty". If the
-- current tag is "TH" or "TD", then that tag is closed and the table tag is set to "TR" to
-- indicate that a row is currently open.
--
-- If the current table tag is "empty", then nothing happens. That means that it is safe to
-- call this function twice in a row to make sure that any current partial row in the table
-- is closed.
function html.table:private_closeTableTags ()
    -- Get the size of the list of tables. if it's 0, there is nothing to do, so return.
    local size = table.getn (self.__tables)
    if size == 0 then return end

    -- Get the tag type of the last tag.
    local tagType = self.__tables[size]

    -- If the tag is TR, then a row with no open tag in it is open, so just close
    -- the row and set the table to the "empty" state to indicate that.
    if tagType == "TR" then
        html:addText ("</TR>\n")
        self.__tables[size] = "empty"
    elseif tagType ~= "empty" then
        -- Here the tag is not TR and not empty, that means that it is one of TD or TH.
        -- Generate a close tag of the approprate type, then change the type to be TR,
        -- to indicate a row that is not closed but has no open tag.
        html:addText ("</" , tagType , ">\n")
        self.__tables[size] = "TR"
    end
end

-------------------------------------------------------------------------------------------

-- Close the most recently opened table. This will close off cells and rows in the table
-- if they are open. Note however that the table doesn't keep track of how many columns
-- a table should have, so incorrect HTML will be generated if the currently open TD tag
-- isn't the last one needed (for example). All it does is keep track of the last tag used
-- for table structure and use that to close it.
function html.table:close ()
    -- Leave if someone tries to close a table when one isn't open.
    if  table.getn (self.__tables) == 0 then return end

    -- Close any open row that might exist.
    html.table:closeRow ()

    -- Close the table tag now.
    html:addText ("</TABLE>\n")

    -- Remove the last item from the top of the table stack.
    table.remove (self.__tables)
end

-------------------------------------------------------------------------------------------

-- If there is a table open, then generate a caption tag for the table. caption is the actual
-- caption text, and align (optional) can be one of LEFT, RIGHT, TOP or BOTTOM.
function html.table:caption (caption , align)
    -- Leave if someone tries to caption a table when one isn't open.
    if  table.getn (self.__tables) == 0 then return end

    html:addText ('<CAPTION ')
    attribute ("ALIGN" , align)
    html:addText (">" , caption , "</CAPTION>\n")
end

-------------------------------------------------------------------------------------------

-- Start a new row in the table. If there is currently a row open, it will be closed (as will
-- any cell currently open) before the new row opens.
function html.table:newRow (height, align, valign, bgcolor, bordercolor)
    -- Get the size of the list of tables. if it's 0, there is nothing to do, so return.
    local size = table.getn (self.__tables)
    if size == 0 then return end

    -- Close any existing row, then open a new one.
    html.table:closeRow ()
    html:addText ("<TR")
    attribute ("HEIGHT" , height)
    attribute ("ALIGN" , align)
    attribute ("VALIGN" , valign)
    attribute ("BGCOLOR" , bgcolor)
    attribute ("BORDERCOLOR" , bordercolor)
    html:addText (">")

    -- Need to make the item at the top of the table stack think that it's got an open
    -- row now.
    self.__tables[size] = "TR"
end

-------------------------------------------------------------------------------------------

-- Close a currently open row. If any cell tags are currently open in the row, then close it
-- before closing the row.
function html.table:closeRow ()
    -- Call the private function twice. The first will close either the current row for us,
    -- or the tag in the current table. The second call will either clow the row or do nothing,
    -- depending what the first did. Either way, the table tag is set to empty, which is what
    -- we want for this.
    html.table:private_closeTableTags ()
    html.table:private_closeTableTags ()
end

-------------------------------------------------------------------------------------------

-- Start a table cell heading. If there is not currently a row open, one will be opened
-- for you. If there is currently a cell open, it will be closed before the new one is opened.
function html.table:heading (nowrap, width, height, align, valign, rowSpan, colSpan, bgcolor, bordercolor)
    -- Get the size of the list of tables. if it's 0, there is nothing to do, so return.
    local size = table.getn (self.__tables)
    if size == 0 then return end

    -- If the tag at the top of the table is not empty and not TR, then there is a tag open
    -- so close it.
    if self.__tables[size] ~= "empty" and self.__tables[size] ~= "TR" then html.table:private_closeTableTags () end

    -- If the tag at the top of the stack is "empty", there is no row open, so open one.
    if self.__tables[size] == "empty" then html.table:newRow () end

    -- Output the proper TH tag, and also change the tag type.
    html:addText ("<TH")
    if nowrap then html:addText (" NOWRAP") end
    attribute ("WIDTH" , width)
    attribute ("HEIGHT" , height)
    attribute ("ALIGN" , align)
    attribute ("VALIGN" , valign)
    attribute ("ROWSPAN" , rowSpan)
    attribute ("COLSPAN" , colSpan)
    attribute ("BGCOLOR" , bgcolor)
    attribute ("BORDERCOLOR" , bordercolor)
    html:addText (">")
    self.__tables[size] = "TH"
end

-------------------------------------------------------------------------------------------

-- Start a table cell definition. If there is not currently a row open, one will be opened
-- for you. If there is currently any sort of cell open, it will be closed before the new
-- one is opened.
function html.table:cell (nowrap, width, height, align, valign, class, style, rowSpan, colSpan, bgcolor, bordercolor)
    -- Get the size of the list of tables. if it's 0, there is nothing to do, so return.
    local size = table.getn (self.__tables)
    if size == 0 then return end

    -- If the tag at the top of the table is not empty and not TR, then there is a tag open
    -- so close it.
    if self.__tables[size] ~= "empty" and self.__tables[size] ~= "TR" then html.table:private_closeTableTags () end

    -- If the tag at the top of the stack is "empty", there is no row open, so open one.
    if self.__tables[size] == "empty" then html.table:newRow () end

    -- Output the proper TD tag, and also change the tag type.
    html:addText ("<TD")
    if nowrap then html:addText (" NOWRAP") end
    attribute ("WIDTH" , width)
    attribute ("HEIGHT" , height)
    attribute ("ALIGN" , align)
    attribute ("VALIGN" , valign)
    attribute ("ROWSPAN" , rowSpan)
    attribute ("COLSPAN" , colSpan)
    attribute ("BGCOLOR" , bgcolor)
    attribute ("BORDERCOLOR" , bordercolor)
    attribute ("CLASS" , class)
    attribute ("STYLE" , style)
    html:addText (">")
    self.__tables[size] = "TD"
end

-------------------------------------------------------------------------------------------

-- Open a new form. Each parameter works the way you would expect. Anything passed in as
-- nil will not be put into the form tag.
function html.form:open (action, method, encoding, target,onSubmit)
    -- Leave if someone tries to open a form when one is already open
    if self.__formopen ~= "no" then return end

    -- Now there is a form open
    self.__formopen = "yes"

    html:addText ("<FORM")

    attribute ("ACTION" , action)
    attribute ("METHOD" , method)
    attribute ("ENCTYPE" , encoding)
    attribute ("TARGET" , target)
    attribute ("ONSUBMIT" , onSubmit)
    html:addText (">\n")
end

-------------------------------------------------------------------------------------------

-- Close the most recently opened form. 
function html.form:close ()
    -- Leave if someone tries to close a form when one isn't open.
    if self.__formopen == "no" then return end

    -- Close the form tag now.
    html:addText ("</FORM>\n")

    -- No form open now
    self.__formopen = "no"
end

-------------------------------------------------------------------------------------------

-- Generate a form checkbox. When the checkbox is checked, the name/value pair is sent in
-- the query to the server. if isChecked is non-nil, then this item is initially checked in
-- the form.
function html.form:checkbox (name, value, isChecked,class,style)
    -- Leave if there is no form open
    if self.__formopen == "no" then return end

    html:addText ('<INPUT TYPE="CHECKBOX"')
    attribute ("NAME" , name)
    attribute ("VALUE" , value)
    attribute ("CLASS" , class)
    attribute ("STYLE" , style)
    if isChecked then html:addText (" CHECKED") end
    html:addText (">\n")
end

-------------------------------------------------------------------------------------------

-- Generate a form radio button. There can be one or more of these. All radio buttons with
-- the same name will become a group that is mutally exclusive. If one is selected, the value
-- associated with it will be returned, along with the name. If isChcked is non-nil, this item
-- is initially selected in the list.
function html.form:radio(name, value, isChecked,class,style)
    -- Leave if there is no form open
    if self.__formopen == "no" then return end

    html:addText ('<INPUT TYPE="RADIO"')
    attribute ("NAME" , name)
    attribute ("VALUE" , value)
    attribute ("CLASS" , class)
    attribute ("STYLE" , style)
    if isChecked then html:addText (" CHECKED") end
    html:addText (">\n")
end

-------------------------------------------------------------------------------------------

-- Generate a hidden form element. Nothing is displayed, but the given name/value pair is
-- returned back when the form is submitted. This can be used, for example, to chain forms
-- together in such a way that the form handler knows what was entered in previous forms.
function html.form:hidden(name, value)
    -- Leave if there is no form open
    if self.__formopen == "no" then return end

    html:addText ('<INPUT TYPE="HIDDEN"')
    attribute ("NAME" , name)
    attribute ("VALUE" , value)
    html:addText (">\n")
end

-------------------------------------------------------------------------------------------

-- Generate a reset button. When this is clicked, the browser will reset the form to it's
-- initial state. This means that any text entered is removed, all radio/checkbox items go
-- to their default state, etc.
--
-- The caption provided will be the text of the button. This won't get submitted back to you.
function html.form:reset(caption,class,style)
    -- Leave if there is no form open
    if self.__formopen == "no" then return end

    html:addText ('<INPUT TYPE="RESET"')
    attribute ("VALUE" , caption)
    attribute ("CLASS" , class)
    attribute ("STYLE" , style)
    html:addText (">\n")
end

-------------------------------------------------------------------------------------------

-- Generate a submit button. The value given will be the caption on the button. When the 
-- user clicks the button, the form is submitted. In the submitted name/value pairs, the 
-- name/value of the submit button is also given. So you can have several submit buttons
-- with the same name but different captions, and distinguish between which one was pressed.
function html.form:submit(name, value,class,style)
    -- Leave if there is no form open
    if self.__formopen == "no" then return end

    html:addText ('<INPUT TYPE="SUBMIT"')
    attribute ("NAME" , name)
    attribute ("VALUE" , value)
    attribute ("CLASS" , class)
    attribute ("STYLE" , style)
    html:addText (">\n")
end

-------------------------------------------------------------------------------------------

-- Generate a submit button that is an image. Clicking on the image submits the form as
-- per usual. In addition, extra fields name.x and name.y are submitted to give you the
-- co-ordinates of where in the image the user clicked to submit the form.
function html.form:image(name, value,src)
    -- Leave if there is no form open
    if self.__formopen == "no" then return end

    html:addText ('<INPUT TYPE="IMAGE"')
    attribute ("NAME" , name)
    attribute ("VALUE" , value)
    attribute ("SRC" , src)
    html:addText (">\n")
end

-------------------------------------------------------------------------------------------

-- Generate a text field in the form, containing the given initial text. fieldSize is how
-- many characters wide the field should be displayed as. maxLength is the maximum length
-- the browser should allow the user to enter. When submitted, name=text comes back to you.
function html.form:text(name, initialText, fieldSize, maxLength,class,style)
    -- Leave if there is no form open
    if self.__formopen == "no" then return end

    html:addText ('<INPUT TYPE="TEXT"')
    attribute ("NAME" , name)
    attribute ("VALUE" , initialText)
    attribute ("SIZE" , fieldSize)
    attribute ("MAXLENGTH" , maxLength)
    attribute ("CLASS" , class)
    attribute ("STYLE" , style)
    html:addText (">\n")
end

-------------------------------------------------------------------------------------------

-- This works like a text field, except the text in the box is not echoed (you see asterixes
-- instead), to help hide what you're typing.
--
-- NOTE: Any password so entered is sent across in plaintext, so this isn't secure unless you
-- are talking over an encrypted connection.
function html.form:password(name, initialText, fieldSize, maxLength,class,style)
    -- Leave if there is no form open
    if self.__formopen == "no" then return end

    html:addText ('<INPUT TYPE="PASSWORD"')
    attribute ("NAME" , name)
    attribute ("VALUE" , initialText)
    attribute ("SIZE" , fieldSize)
    attribute ("MAXLENGTH" , maxLength)
    attribute ("CLASS" , class)
    attribute ("STYLE" , style)
    html:addText (">\n")
end

-------------------------------------------------------------------------------------------

-- Open a text area. When submitted, all text typed in will be returned using the given
-- name. cols and rows give the size of the text area, and wrap determines how word wrapping
-- is done.
--
-- When you use this function, any text output between this and a call to closeTextArea
-- will be the default text. Note also that it is up to you to close the text area before
-- you close the form, the system won't do it for you.
function html.form:openTextArea (name, cols, rows, wrap,class,style)
    -- Leave if there is no form open
    if self.__formopen == "no" then return end

    html:addText ("<TEXTAREA")
    attribute ("NAME" , name)
    attribute ("COLS" , cols)
    attribute ("ROWS" , rows)
    attribute ("WRAP" , wrap)
    attribute ("CLASS" , class)
    attribute ("STYLE" , style)
    html:addText (">\n")
end

-------------------------------------------------------------------------------------------

-- Closes a previously opened text area.
function html.form:closeTextArea ()
    -- Leave if there is no form open
    if self.__formopen == "no" then return end

    html:addText ("</TEXTAREA>")
end

-------------------------------------------------------------------------------------------

-- Generate a complete text area tag in one action. This will generate the text area
-- using the given text, then close it immediately.
function html.form:textArea(name, text, cols, rows, wrap,class,style)
    -- Leave if there is no form open
    if self.__formopen == "no" then return end

    html.form:openTextArea (name , cols, rows, wrap,class,style)
    html:addText (text)
    html.form:closeTextArea ()
end

-------------------------------------------------------------------------------------------

-- Open a select tag. When submitted, all items selected in the select tag will be returned
-- using the name given here and the value associated with the select option. If given,
-- size specifies how many items should be visible at one time. When set to 1 (default), you 
-- get a dropdown list. With > 1, you get a list box. if multipleSelect is true, then multiple
-- items can be selected.
--
-- It is up to the caller to output one or more option tags between this and a call to
-- closeSelect. It is also up to you to call closeSelect manually before closing the form,
-- the system won't do it for you.
function html.form:openSelect (name, size, multipleSelect, class, style)
    -- Leave if there is no form open
    if self.__formopen == "no" then return end

    html:addText ("<SELECT")
    attribute ("NAME" , name)
    attribute ("SIZE" , size)
    attribute ("CLASS" , class)
    attribute ("STYLE" , style)
    if multipleSelect then html:addText (" MULTIPLE") end
    html:addText (">\n")
end

-------------------------------------------------------------------------------------------

-- Closes a previously opened select tag
function html.form:closeSelect ()
    -- Leave if there is no form open
    if self.__formopen == "no" then return end

    html:addText ("</SELECT>")
end

-------------------------------------------------------------------------------------------

-- Output an option tag to the current form. text is what will be visible in the page,
-- and value will be what is returned if that item is selected when the form is submitted.
-- If isSelected is non-nil, then this item will be selected when the page is first viewed.
function html.form:selectOption (text, value, isSelected)
    -- Leave if there is no form open
    if self.__formopen == "no" then return end

    html:addText ("<OPTION")
    attribute ("VALUE" , value)
    if isSelected then html:addText (" SELECTED") end
    html:addText (">" , text , "\n")
end

-------------------------------------------------------------------------------------------

-- Generate a select tag all in one shot. The first three parameters are the same as a
-- call to openSelect. functor should be a function which takes two parameters. It will be
-- called as functor (items, userInfo). This function should generate one or more option
-- tags using selectOption () before returning. After functor returns, the select tag will
-- be closed.
function html.form:select (name, size, multipleSelect, class, style, functor, items, userInfo)
    -- Leave if there is no form open
    if self.__formopen == "no" then return end

    -- Open the select tag, output items, then close it.
    html.form:openSelect (name, size, multipleSelect, class, style)
    functor (items, userInfo)
    html.form:closeSelect ()
end

-------------------------------------------------------------------------------------------

-- Add a file upload component to a form. In order for this to work, the encoding type
-- of the form must be set to "multipart/form-data". Otherwise, nothing will happen.
--
-- This allows someone to upload a file. The results will end up in CGI.parsedMultipartData
-- at runtime, where name is the name given here, and value is a binary string which contains
-- the contents of the file.
function html.form:file (name, class, style)
    -- Leave if there is no form open
    if self.__formopen == "no" then return end

    html:addText ('<INPUT TYPE="FILE"')
    attribute ("NAME" , name)
    attribute ("CLASS" , class)
    attribute ("STYLE" , style)
    html:addText (">\n")
end

-------------------------------------------------------------------------------------------

-- This generates a form with a single submit button containing the given text, which will
-- submit to the url provided with 2 form components. One named "Button" whose value is the
-- button text provided, and one named "command", whose value is the command you pass in.
--
-- method should be "POST" unless the URL you are submitting to isn't a script. In that
-- case, use "GET".
--
-- This won't do anything if there is currently a form open.
function html.form:singleButtonForm (buttonText, command, url, method,encoding,target,class,style)
    -- Leave if there is a form open
    if self.__formopen ~= "no" then return end

    -- Open the form
    html.form:open (url, method, encoding, target)

    -- Put in a submit button and a hidden item
    html.form:submit ("Button" , buttonText,class,style)
    html.form:hidden ("command" , command)

    -- All done, close the form now.
    html.form:close ()
end

-------------------------------------------------------------------------------------------

